#include <qpushbutton.h>
#include <qcombobox.h>
#include <qradiobutton.h>
#include <qapplication.h>
#include <qcheckbox.h>
#include <qlineedit.h>
#include <qspinbox.h>

#include <stdlib.h>  // for qsort()

#include "../../../Core/Scene/src/Scene.h"
#include "../../../Core/Scene/src/Object.h"
#include "../../../Core/DataStructures/src/NeighbourHood.h"
#include "../../../Core/DataStructures/src/SurfelCollection.h"
#include "../../../Core/CoreTools/src/CoreTools.h"
#include "../../../Core/CoreResourceTools/src/CoreResourceTools.h"
#include "../../../Core/Brush/src/Brush.h"
#include "../../../Core/Painter/src/Painter.h"

#include "../../../Utilities/UserMessages/src/StatusBar.h"
#include "../../../Utilities/src/Matrix.h"

#include "Creator_DialogImpl.h"





// ******
// global
// ******


/* comparison routine for qsort */
int compare (const void *u, const void *v)  {
	return (*(Vector3D*) u)[0] <= (*(Vector3D*)v)[0] ? -1 : 1;
}



// the single Creator_DialogImpl instance, only
// accesible by getInstance() and
// destroyInstance()
static Creator_DialogImpl *instance = 0;

// *******************
// private c'tor/d'tor
// *******************

Creator_DialogImpl::Creator_DialogImpl (QWidget *parent, const char *name, bool modal, WFlags flags)
    : Creator_DialogBase (parent, name, modal, flags) {


}

Creator_DialogImpl::~Creator_DialogImpl() {
    // no need to delete child widgets, Qt does it all for us
}

// **************
// public methods
// **************

Creator_DialogImpl *Creator_DialogImpl::getInstance() {

	if (instance == 0) {
		instance = new Creator_DialogImpl();
	}
	return instance;

}

void Creator_DialogImpl::destroyInstance() {

	if (instance != 0) {
		instance->close();
		delete instance;
	}
	instance = 0;

}

// ***************
// private methods
// ***************







void Creator_DialogImpl::handleCreateSphere() {

	Scene            *scene;
	Object			 *newObject;
	SurfelCollection *surfels;
	SurfelInterface  *newSurfel;
	Vector3D		 newPosition,
					 newNormal,
					 newAxisOne,
					 newAxisTwo;
	int				 nofPoints;
	int				 subdivisionDepth;
	int				 quadCount;
	Quad			 *quad,
					 *newQuad;
	Vector3D		 edgeAB,
					 edgeBC,
					 edgeCD,
					 edgeDA,
					 center;
	int				 i,
					 j,
					 nofFinalPoints;


	QApplication::setOverrideCursor (Qt::waitCursor);
	
	scene = Scene::getInstance();
	

	subdivisionDepth = this->subdivisionLevelsSpinBox->value();


	// create cube	
	quad = new Quad();
	quad->recursion = subdivisionDepth;
	quad->a = Vector3D ( 1,  1,  1);
	quad->b = Vector3D (-1,  1,  1);
	quad->c = Vector3D (-1, -1,  1);
	quad->d = Vector3D ( 1, -1,  1);
	quad->a.normalize();
	quad->b.normalize();
	quad->c.normalize();
	quad->d.normalize();
	quad->coords[0] = quad->coords[1] = 0.0f;
	quad->coords[2] = quad->coords[3] = 1.0f;
	mesh.append(quad);

	quad = new Quad();
	quad->recursion = subdivisionDepth;
	quad->a = Vector3D ( 1,  1,  1);
	quad->b = Vector3D (-1,  1,  1);
	quad->c = Vector3D (-1,  1,  -1);
	quad->d = Vector3D ( 1,  1,  -1);
	quad->a.normalize();
	quad->b.normalize();
	quad->c.normalize();
	quad->d.normalize();
	quad->coords[0] = quad->coords[1] = 0.0f;
	quad->coords[2] = quad->coords[3] = 1.0f;
	mesh.append(quad);

	quad = new Quad();
	quad->recursion = subdivisionDepth;
	quad->a = Vector3D ( 1,  1,  1);
	quad->b = Vector3D ( 1, -1,  1);
	quad->c = Vector3D ( 1, -1,  -1);
	quad->d = Vector3D ( 1,  1,  -1);
	quad->a.normalize();
	quad->b.normalize();
	quad->c.normalize();
	quad->d.normalize();
	quad->coords[0] = quad->coords[1] = 0.0f;
	quad->coords[2] = quad->coords[3] = 1.0f;
	mesh.append(quad);

	quad = new Quad();
	quad->recursion = subdivisionDepth;
	quad->a = Vector3D ( 1,  1, -1);
	quad->b = Vector3D (-1,  1, -1);
	quad->c = Vector3D (-1, -1, -1);
	quad->d = Vector3D ( 1, -1, -1);
	quad->a.normalize();
	quad->b.normalize();
	quad->c.normalize();
	quad->d.normalize();
	quad->coords[0] = quad->coords[1] = 0.0f;
	quad->coords[2] = quad->coords[3] = 1.0f;
	mesh.append(quad);

	quad = new Quad();
	quad->recursion = subdivisionDepth;
	quad->a = Vector3D ( 1, -1,  1);
	quad->b = Vector3D (-1, -1,  1);
	quad->c = Vector3D (-1, -1,  -1);
	quad->d = Vector3D ( 1, -1,  -1);
	quad->a.normalize();
	quad->b.normalize();
	quad->c.normalize();
	quad->d.normalize();
	quad->coords[0] = quad->coords[1] = 0.0f;
	quad->coords[2] = quad->coords[3] = 1.0f;
	mesh.append(quad);

	quad = new Quad();
	quad->recursion = subdivisionDepth;
	quad->a = Vector3D (-1,  1,  1);
	quad->b = Vector3D (-1, -1,  1);
	quad->c = Vector3D (-1, -1,  -1);
	quad->d = Vector3D (-1,  1,  -1);
	quad->a.normalize();
	quad->b.normalize();
	quad->c.normalize();
	quad->d.normalize();
	quad->coords[0] = quad->coords[1] = 0.0f;
	quad->coords[2] = quad->coords[3] = 1.0f;
	mesh.append(quad);

	quadCount = 6;


	while (quadCount > 0) {

		quad = mesh.first();

		if (quad->recursion > 0) {
			
			mesh.remove();

			edgeAB = quad->a + quad->b;
			edgeAB.normalize();
			edgeBC = quad->b + quad->c;
			edgeBC.normalize();
			edgeCD = quad->c + quad->d;
			edgeCD.normalize();
			edgeDA = quad->d + quad->a;
			edgeDA.normalize();
			center = quad->a + quad->b + quad->c + quad->d;
			center.normalize();

			newQuad = new Quad();
			newQuad->recursion = quad->recursion - 1;
			newQuad->a = quad->a;
			newQuad->b = edgeAB;
			newQuad->c = center;
			newQuad->d = edgeDA;
			newQuad->coords[0] = quad->coords[0];
			newQuad->coords[1] = quad->coords[1];
			newQuad->coords[2] = 0.5f * (quad->coords[0] + quad->coords[2]);
			newQuad->coords[3] = 0.5f * (quad->coords[1] + quad->coords[3]);
			mesh.append(newQuad);

			newQuad = new Quad();
			newQuad->recursion = quad->recursion - 1;
			newQuad->a = edgeAB;
			newQuad->b = quad->b;
			newQuad->c = edgeBC;
			newQuad->d = center;
			newQuad->coords[0] = 0.5f * (quad->coords[0] + quad->coords[2]);
			newQuad->coords[1] = quad->coords[1];
			newQuad->coords[2] = quad->coords[2];
			newQuad->coords[3] = 0.5f * (quad->coords[1] + quad->coords[3]);
			mesh.append(newQuad);

			newQuad = new Quad();
			newQuad->recursion = quad->recursion - 1;
			newQuad->a = center;
			newQuad->b = edgeBC;
			newQuad->c = quad->c;
			newQuad->d = edgeCD;
			newQuad->coords[0] = 0.5f * (quad->coords[0] + quad->coords[2]);
			newQuad->coords[1] = 0.5f * (quad->coords[1] + quad->coords[3]);
			newQuad->coords[2] = quad->coords[2];
			newQuad->coords[3] = quad->coords[3];
			mesh.append(newQuad);

			newQuad = new Quad();
			newQuad->recursion = quad->recursion - 1;
			newQuad->a = edgeDA;
			newQuad->b = center;
			newQuad->c = edgeCD;
			newQuad->d = quad->d;
			newQuad->coords[0] = quad->coords[0];
			newQuad->coords[1] = 0.5f * (quad->coords[1] + quad->coords[3]);
			newQuad->coords[2] = 0.5f * (quad->coords[0] + quad->coords[2]);
			newQuad->coords[3] = quad->coords[3];
			mesh.append(newQuad);

			quadCount += 4;

			delete quad;
		}
		else {
			quadCount--;
		}
	}


	
	newObject = new Object("Sphere");
	surfels = newObject->getSurfelCollection();
	SurfelInterface::PropertyDescriptor pd = SurfelInterface::NORMAL | SurfelInterface::TANGENT_AXES | SurfelInterface::POSITION | SurfelInterface::FLAGS;;
	surfels->setPropertyDescriptor(pd);
	
				
	nofPoints = 4 * mesh.count();	
	
	Vector3D *newPos = new Vector3D [nofPoints];
	float    *newU   = new float [nofPoints];
	float    *newV   = new float [nofPoints];
	bool     *skip   = new bool [nofPoints];
	i = 0;
	while (mesh.count() > 0) {

		quad = mesh.first();
		mesh.remove();

		newPos[i] = quad->a;
		newU[i] = quad->coords[0];
		newV[i] = quad->coords[1];
		skip[i++] = false;
		newPos[i] = quad->b;
		newU[i] = quad->coords[2];
		newV[i] = quad->coords[1];
		skip[i++] = false;
		newPos[i] = quad->c;
		newU[i] = quad->coords[2];
		newV[i] = quad->coords[3];
		skip[i++] = false;
		newPos[i] = quad->d;
		newU[i] = quad->coords[0];
		newV[i] = quad->coords[3];
		skip[i++] = false;		
		
		delete quad;
	}

	


	float bumpsScale = 0.01f * (float) this->bumpsScaleSpinBox->value();
	float bumpsFrequency = this->bumpsFrequencySpinBox->value();

	Vector3D  displPos;

	for (i = 0; i < nofPoints; i++) {
		
		newPos[i] *= 100.0f * (1 + bumpsScale * cos (newU[i] * 2.0f * M_PI * bumpsFrequency) * cos (newV[i] * 2.0f * M_PI * bumpsFrequency));
	}

	qsort (newPos, nofPoints, sizeof newPos[0], compare);


	for (i = 0; i < nofPoints; i++) {
		if (!skip[i]) {

			j = i + 1;
			while ((newPos[j][0] == newPos[i][0]) && (j < nofPoints)) {
				if ((newPos[j][1] == newPos[i][1]) && (newPos[j][2] == newPos[i][2])) {
					skip[j] = true;
				}
				j++;
			}
		}
	}


	
	nofFinalPoints = 0;
	for (i = 0; i < nofPoints; i++) {
		if (!skip[i]) {
			nofFinalPoints++;
		}
	}

	Vector3D *finalPos = new Vector3D [nofFinalPoints];
	Vector3D *finalNormals = new Vector3D [nofFinalPoints];

	nofFinalPoints = 0;
	for (i = 0; i < nofPoints; i++) {
		if (!skip[i]) {
			finalPos[nofFinalPoints] = newPos[i];
			nofFinalPoints++;
		}
	}


	NeighbourHood *search = new NeighbourHood (finalPos, nofFinalPoints);
	search->setNofNeighbours (20);
	for (i = 0; i < nofFinalPoints; i++) {

		search->setSourcePoint (finalPos[i]);
		finalNormals[i] = search->computeNormal ();
		if (Vector3D::dotProduct (finalNormals[i], finalPos[i]) < 0) {
			finalNormals[i] = -finalNormals[i];
		}
	}

	surfels->reserve (nofFinalPoints);
		
	
	for (i = 0; i < nofFinalPoints; i++) {
		
		newSurfel = surfels->addSurfel(false);
		
		finalPos[i][2] += 35*5;
		scene->setSurfelPosition (newSurfel, 0.2f * finalPos[i], false);

		finalNormals[i].getTwoOrthogonals (newAxisOne, newAxisTwo);
		newAxisOne.normalize();
		newAxisTwo.normalize();

		scene->setSurfelNormalTangentAxesSystem (newSurfel, finalNormals[i], newAxisOne, newAxisTwo, false);
		
	}	

	delete[] newPos;
	delete[] newU;
	delete[] newV;
	delete[] skip;


	scene->addObject(newObject);
	
	QApplication::restoreOverrideCursor ();
}





void Creator_DialogImpl::handleCreateEdge() {
	
	Scene            *scene;
	Object			 *newObject;
	SurfelCollection *surfels;
	SurfelInterface  *newSurfel;
	Vector3D		 newPosition,
					 newNormal,
					 newAxisOne,
					 newAxisTwo;
	int				 i,
					 j;
	float			 sphereRadius;	
	int				 nofRoundedPoints = this->edgeRoundPointsSpinBox->value();
	int				 nofStraightPoints = this->edgeStraightPointsSpinBox->value();
	int				 multiplicity = this->edgeMultiplicitySpinBox->value();
	
				
	QApplication::setOverrideCursor (Qt::waitCursor);
	
	scene = Scene::getInstance();
		
	sphereRadius = 2.0f /  MY_PI * nofRoundedPoints;
		
	newObject = new Object("Edge");
	surfels = newObject->getSurfelCollection();
	SurfelInterface::PropertyDescriptor pd = SurfelInterface::NORMAL | SurfelInterface::TANGENT_AXES | SurfelInterface::POSITION
		| SurfelInterface::FLAGS;
	surfels->setPropertyDescriptor(pd);
		
	surfels->reserve ((nofRoundedPoints + 2 * nofStraightPoints) * multiplicity);
	
	
	for (j = 0; j < multiplicity; j++) {
		for (i = 0; i < nofRoundedPoints; i++) {
			
			newSurfel = surfels->addSurfel(false);
			
			newNormal[0] = sin (0.5f * MY_PI * (float) i / (float) (nofRoundedPoints - 1));
			newNormal[1] = cos (0.5f * MY_PI * (float) i / (float) (nofRoundedPoints - 1));
			newNormal[2] = 0;
			
			newPosition = sphereRadius * newNormal;
			newPosition[2] = j;
			
			newNormal.getTwoOrthogonals (newAxisOne, newAxisTwo);
			newAxisOne.normalize();
			newAxisTwo.normalize();

			scene->setSurfelPosition (newSurfel, newPosition, false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, newNormal, newAxisOne, newAxisTwo, false);
		}		
		
		
		
		for (i = 0; i < nofStraightPoints; i++) {
			
			newSurfel = surfels->addSurfel(false);
			
			
			scene->setSurfelPosition (newSurfel, Vector3D (-i, sphereRadius, j), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (0, 1, 0), Vector3D (1, 0, 0), Vector3D (0, 0, -1), false);
						
			newSurfel = surfels->addSurfel(false);
			
			scene->setSurfelPosition (newSurfel, Vector3D (sphereRadius, -i, j), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (1, 0, 0), Vector3D (0, 1, 0), Vector3D (0, 0, -1), false);
		
			}
	}	

	scene->addObject(newObject);
	
	QApplication::restoreOverrideCursor ();
}





void Creator_DialogImpl::handleCreatePlane() {
	
	Scene            *scene;
	Object			 *newObject;
	SurfelCollection *surfels;
	SurfelInterface  *newSurfel;
	int				 i,
					 j;
	int				 nofPointsX = this->planeXPointsSpinBox->value();
	int				 nofPointsY = this->planeYPointsSpinBox->value();
	
				
	QApplication::setOverrideCursor (Qt::waitCursor);
	
	scene = Scene::getInstance();
		
		
	newObject = new Object("Plane");
	surfels = newObject->getSurfelCollection();
	SurfelInterface::PropertyDescriptor pd = SurfelInterface::NORMAL | SurfelInterface::TANGENT_AXES | SurfelInterface::POSITION
		| SurfelInterface::FLAGS;
	surfels->setPropertyDescriptor(pd);
		
	surfels->reserve (nofPointsX * nofPointsY);
	
	
	for (j = 0; j < nofPointsY; j++) {
		for (i = 0; i < nofPointsX; i++) {
			
			newSurfel = surfels->addSurfel(false);

			scene->setSurfelPosition (newSurfel, Vector3D (i - nofPointsX/2, j - nofPointsY/2, 0), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (0, 0, 1), Vector3D (1, 0, 0), Vector3D (0, 1, 0), false);
		}						
	}	
	
	scene->addObject(newObject);
	
	QApplication::restoreOverrideCursor ();
}






void Creator_DialogImpl::handleCreateCube() {
	
	Scene            *scene;
	Object			 *newObject;
	SurfelCollection *surfels;
	SurfelInterface  *newSurfel;
	int				 i,
					 j;
	float			 nofPointsX = this->cubeXPointsSpinBox->value();
	float			 nofPointsY = this->cubeYPointsSpinBox->value();
	float			 nofPointsZ = this->cubeZPointsSpinBox->value();
	
				
	QApplication::setOverrideCursor (Qt::waitCursor);
	
	scene = Scene::getInstance();
		
		
	newObject = new Object("Block");
	surfels = newObject->getSurfelCollection();
	SurfelInterface::PropertyDescriptor pd = SurfelInterface::NORMAL | SurfelInterface::TANGENT_AXES | SurfelInterface::POSITION
		| SurfelInterface::FLAGS;
	surfels->setPropertyDescriptor(pd);
		
	surfels->reserve (2*nofPointsX*nofPointsY + 2*nofPointsX*nofPointsZ + 2*nofPointsY*nofPointsZ);
	
	
	for (j = 0; j < nofPointsY; j++) {
		for (i = 0; i < nofPointsX; i++) {			
			newSurfel = surfels->addSurfel(false);
			scene->setSurfelPosition (newSurfel, Vector3D (0.5 + i - nofPointsX/2, 0.5 + j - nofPointsY/2, nofPointsZ/2), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (0, 0, 1), Vector3D (1, 0, 0), Vector3D (0, 1, 0), false);
			newSurfel = surfels->addSurfel(false);
			scene->setSurfelPosition (newSurfel, Vector3D (0.5 + i - nofPointsX/2, 0.5 + j - nofPointsY/2, -nofPointsZ/2), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (0, 0, -1), Vector3D (1, 0, 0), Vector3D (0, -1, 0), false);
		}						
	}	

		
	for (j = 0; j < nofPointsZ; j++) {
		for (i = 0; i < nofPointsX; i++) {			
			newSurfel = surfels->addSurfel(false);
			scene->setSurfelPosition (newSurfel, Vector3D (0.5 + i - nofPointsX/2, nofPointsY/2, 0.5 + j - nofPointsZ/2 ), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (0, 1, 0), Vector3D (0, 0, 1), Vector3D (1, 0, 0), false);
			newSurfel = surfels->addSurfel(false);
			scene->setSurfelPosition (newSurfel, Vector3D (0.5 + i - nofPointsX/2, -nofPointsY/2, 0.5 + j - nofPointsZ/2 ), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (0, -1, 0), Vector3D (0, 0, 1), Vector3D (-1, 0, 0), false);
		}						
	}	
	
	for (j = 0; j < nofPointsZ; j++) {
		for (i = 0; i < nofPointsY; i++) {			
			newSurfel = surfels->addSurfel(false);
			scene->setSurfelPosition (newSurfel, Vector3D (nofPointsX/2, 0.5 + i - nofPointsY/2, 0.5 + j - nofPointsZ/2 ), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (1, 0, 0), Vector3D (0, 1, 0), Vector3D (0, 0, 1), false);
			newSurfel = surfels->addSurfel(false);
			scene->setSurfelPosition (newSurfel, Vector3D (-nofPointsX/2, 0.5 + i - nofPointsY/2, 0.5 + j - nofPointsZ/2 ), false);
			scene->setSurfelNormalTangentAxesSystem (newSurfel, Vector3D (-1, 0, 0), Vector3D (0, -1, 0), Vector3D (0, 0, 1), false);
		}						
	}	
	
	scene->addObject(newObject);
	
	QApplication::restoreOverrideCursor ();
}




void Creator_DialogImpl::handleCreateCylinder() {
	
	Scene            *scene;
	Object			 *newObject;
	SurfelCollection *surfels;
	SurfelInterface  *newSurfel;
	int				 i,
					 j;
	int				 nofPointsLength = this->cylLengthPointsSpinBox->value();
	int				 nofPointsDiameter = this->cylDiameterPointsSpinBox->value();
	float			 angle,
					 radius;
	Vector3D		 normal,
					 axisOne,
					 axisTwo;			

	QApplication::setOverrideCursor (Qt::waitCursor);
	
	scene = Scene::getInstance();
				
	newObject = new Object("Cylinder");
	surfels = newObject->getSurfelCollection();
	SurfelInterface::PropertyDescriptor pd = SurfelInterface::NORMAL | SurfelInterface::TANGENT_AXES | SurfelInterface::POSITION
		| SurfelInterface::FLAGS;
	surfels->setPropertyDescriptor(pd);
		
	surfels->reserve (nofPointsLength * nofPointsDiameter);
	
	radius = (float) nofPointsDiameter / (2.0f * M_PI);
	
	for (j = 0; j < nofPointsLength; j++) {
		for (i = 0; i < nofPointsDiameter; i++) {			
			newSurfel = surfels->addSurfel(false);
			angle = 2.0f * M_PI * i / nofPointsDiameter;
			scene->setSurfelPosition (newSurfel, Vector3D (radius * sin (angle), radius * cos(angle), j + 0.5f - (float)nofPointsLength / 2), false);
			normal = Vector3D (sin (angle), cos(angle), 0);
			normal.getTwoOrthogonals (axisOne, axisTwo);
			axisOne.normalize();
			axisTwo.normalize();			
			scene->setSurfelNormalTangentAxesSystem (newSurfel, normal, axisOne, axisTwo, false);
		}						
	}	

	
	scene->addObject(newObject);
	
	QApplication::restoreOverrideCursor ();
}



void Creator_DialogImpl::handleCreateFromActive() {
	
	Scene            *scene;
	Object			 *newObject,
					 *activeObject;
	SurfelCollection *surfels,
					 *oldSurfels;
	SurfelInterface  *newSurfel,
					 *oldSurfel;

	QApplication::setOverrideCursor (Qt::waitCursor);
	
	scene = Scene::getInstance();

	activeObject = scene->getActiveObject();
	oldSurfels = activeObject->getSurfelCollection();

	
	newObject = new Object("New Copy");
	surfels = newObject->getSurfelCollection();
	SurfelInterface::PropertyDescriptor pd = SurfelInterface::NORMAL | SurfelInterface::TANGENT_AXES | SurfelInterface::POSITION
		| SurfelInterface::FLAGS;
	surfels->setPropertyDescriptor(pd);
		
	surfels->reserve (oldSurfels->getNofSurfels());
	
	Vector3D normal, axisOne, axisTwo;


	for (oldSurfel = oldSurfels->getFirstSurfel(); oldSurfel; oldSurfel = oldSurfels->getNextSurfel()) {
		
		newSurfel = surfels->addSurfel(false);
		scene->setSurfelPosition (newSurfel, oldSurfel->getPosition(), false);
		normal = oldSurfel->getNormal();
		normal.getTwoOrthogonals (axisOne, axisTwo);
		axisOne.normalize();
		axisTwo.normalize();
		scene->setSurfelNormalTangentAxesSystem (newSurfel, normal, axisOne, axisTwo, false);

	
		Vector3D p = newSurfel->getPosition();
		Vector3D n = newSurfel->getNormal();
		Vector3D a1 = newSurfel->getAxisOne();
		Vector3D a2 = newSurfel->getAxisTwo();
	}	

	scene->addObject(newObject);
	
	QApplication::restoreOverrideCursor ();
}


